from dataclasses import dataclass
from typing import Optional

from opentelemetry import metrics as otel_metrics, trace as otel_trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter

from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter

from .logger import setup_json_logging

@dataclass
class Config:
    service_name: str = "ampy-python"
    service_version: str = "0.1.0"
    environment: str = "dev"
    collector_endpoint: str = "localhost:4317"  # gRPC
    enable_logs: bool = True
    enable_tracing: bool = True
    enable_metrics: bool = True
    sample_always: bool = True

_tracer_provider: Optional[TracerProvider] = None
_meter_provider: Optional[MeterProvider] = None

def init(cfg: Config) -> None:
    resource = Resource.create({
        "service.name": cfg.service_name,
        "service.version": cfg.service_version,
        "deployment.environment": cfg.environment,
    })

    # ----- Logs -----
    if cfg.enable_logs:
        setup_json_logging(cfg.service_name, cfg.environment, cfg.service_version)

    # ----- Tracing -----
    if cfg.enable_tracing:
        span_exporter = OTLPSpanExporter(endpoint=cfg.collector_endpoint, insecure=True)
        sp = BatchSpanProcessor(span_exporter)

        global _tracer_provider
        _tracer_provider = TracerProvider(resource=resource)
        _tracer_provider.add_span_processor(sp)
        if cfg.sample_always:
            # AlwaysOn is the default sampler for Python when not overridden; BatchSpanProcessor handles buffering.
            pass
        otel_trace.set_tracer_provider(_tracer_provider)

    # ----- Metrics -----
    if cfg.enable_metrics:
        metric_exporter = OTLPMetricExporter(endpoint=cfg.collector_endpoint, insecure=True)
        reader = PeriodicExportingMetricReader(metric_exporter)
        global _meter_provider
        _meter_provider = MeterProvider(resource=resource, metric_readers=[reader])
        otel_metrics.set_meter_provider(_meter_provider)

def shutdown() -> None:
    # Flush providers if present
    if _tracer_provider is not None:
        _tracer_provider.shutdown()
    if _meter_provider is not None:
        _meter_provider.shutdown()



from opentelemetry import metrics as otel_metrics
from opentelemetry.metrics import Counter, Histogram
from opentelemetry.sdk.metrics import MeterProvider

# Instruments (module-level singletons after init)
_bus_produced: Counter
_bus_consumed: Counter
_bus_delivery_latency: Histogram
_oms_order_submit: Counter
_oms_order_latency: Histogram
_oms_rejections: Counter

def init_instruments() -> None:
    meter = otel_metrics.get_meter_provider().get_meter("ampyobs")

    global _bus_produced, _bus_consumed, _bus_delivery_latency
    global _oms_order_submit, _oms_order_latency, _oms_rejections

    _bus_produced = meter.create_counter(
        name="ampy.bus.produced_total",
        description="Messages produced to ampy-bus",
    )
    _bus_consumed = meter.create_counter(
        name="ampy.bus.consumed_total",
        description="Messages consumed from ampy-bus",
    )
    _bus_delivery_latency = meter.create_histogram(
        name="ampy.bus.delivery_latency_ms",
        unit="ms",
        description="Bus end-to-end delivery latency (ms)",
    )

    _oms_order_submit = meter.create_counter(
        name="ampy.oms.order_submit_total",
        description="Order submissions by outcome",
    )
    _oms_order_latency = meter.create_histogram(
        name="ampy.oms.order_latency_ms",
        unit="ms",
        description="OMS order latency (submit→ack) in ms",
    )
    _oms_rejections = meter.create_counter(
        name="ampy.oms.rejections_total",
        description="Order rejections by reason",
    )

def bus_produced(topic: str, n: int = 1, service: str | None = None, env: str | None = None) -> None:
    _bus_produced.add(n, {"topic": topic, "service": service or "", "env": env or ""})

def bus_consumed(topic: str, n: int = 1, service: str | None = None, env: str | None = None) -> None:
    _bus_consumed.add(n, {"topic": topic, "service": service or "", "env": env or ""})

def bus_delivery_latency_ms(topic: str, ms: float, service: str | None = None, env: str | None = None) -> None:
    _bus_delivery_latency.record(ms, {"topic": topic, "service": service or "", "env": env or ""})

def oms_order_submit(broker: str, outcome: str, service: str | None = None, env: str | None = None) -> None:
    _oms_order_submit.add(1, {"broker": broker, "outcome": outcome, "service": service or "", "env": env or ""})

def oms_order_latency_ms(broker: str, ms: float, service: str | None = None, env: str | None = None) -> None:
    _oms_order_latency.record(ms, {"broker": broker, "service": service or "", "env": env or ""})

def oms_reject(broker: str, reason: str, service: str | None = None, env: str | None = None) -> None:
    _oms_rejections.add(1, {"broker": broker, "reason": reason, "service": service or "", "env": env or ""})



__all__ = [
    "Config",
    "init",
    "shutdown",
    "logger",
    "metrics",
    "tracing",
    "propagation",
]

from .obs import Config, init, shutdown
from . import logger, metrics, tracing, propagation



import logging
import sys
from typing import Any, Optional

from opentelemetry import trace as otel_trace

# Global logger instance
L: Optional[logging.LoggerAdapter] = None

# Fallback logger initialization
def _get_logger():
    if L is None:
        # Create a basic logger if not initialized
        logger = logging.getLogger("ampyobs")
        if not logger.handlers:
            handler = logging.StreamHandler(sys.stdout)
            handler.setFormatter(JsonFormatter())
            logger.addHandler(handler)
            logger.setLevel(logging.INFO)
        return logger
    return L

class TraceContextFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        span = otel_trace.get_current_span()
        span_ctx = span.get_span_context()
        if span_ctx and span_ctx.is_valid:
            record.trace_id = span_ctx.trace_id
            record.span_id = span_ctx.span_id
            # hex format
            record.trace_id_hex = format(span_ctx.trace_id, "032x")
            record.span_id_hex = format(span_ctx.span_id, "016x")
        else:
            record.trace_id = None
            record.span_id = None
            record.trace_id_hex = None
            record.span_id_hex = None
        return True

class JsonFormatter(logging.Formatter):
    # Minimal JSON without extra deps; relies on default LogRecord.__dict__
    def format(self, record: logging.LogRecord) -> str:
        # Build a dict of stable fields
        d: dict[str, Any] = {
            "ts": self.formatTime(record, "%Y-%m-%dT%H:%M:%S.%fZ"),
            "level": record.levelname.lower(),
            "message": record.getMessage(),
            "service": getattr(record, "service", None),
            "env": getattr(record, "env", None),
            "service_version": getattr(record, "service_version", None),
            "trace_id": getattr(record, "trace_id_hex", None),
            "span_id": getattr(record, "span_id_hex", None),
        }
        # Include any extra attributes from `logger.bind(...)` style calls
        for k, v in record.__dict__.items():
            if k in d or k.startswith("_"):
                continue
            if k in ("args", "msg", "exc_text", "exc_info", "stack_info", "lineno", "pathname", "filename", "module", "funcName", "created", "msecs", "relativeCreated", "thread", "threadName", "process", "processName"):
                continue
            d[k] = v
        import json
        return json.dumps(d, separators=(",", ":"), ensure_ascii=False)

def setup_json_logging(service: str, env: str, version: str) -> None:
    root = logging.getLogger()
    root.setLevel(logging.INFO)

    handler = logging.StreamHandler(sys.stdout)
    fmt = JsonFormatter()
    handler.setFormatter(fmt)

    # Attach default fields to every record
    class ServiceAdapter(logging.LoggerAdapter):
        def process(self, msg, kwargs):
            extra = kwargs.setdefault("extra", {})
            extra.setdefault("service", service)
            extra.setdefault("env", env)
            extra.setdefault("service_version", version)
            return msg, kwargs

    # Add trace filter
    handler.addFilter(TraceContextFilter())
    root.handlers = [handler]

    # Expose a module-level logger adapter
    global L
    L = ServiceAdapter(logging.getLogger("ampyobs"), {})
    
    # Also set up a fallback logger if L is still None
    if L is None:
        L = ServiceAdapter(logging.getLogger("ampyobs"), {})



from typing import Dict, Optional
from opentelemetry.context import get_current
from opentelemetry.propagate import get_global_textmap

def inject_trace(headers: Dict[str, str], context=None) -> None:
    ctx = context or get_current()
    get_global_textmap().inject(headers, context=ctx)

def extract_trace(headers: Dict[str, str]):
    return get_global_textmap().extract(headers)



from dataclasses import dataclass
from contextlib import contextmanager
from typing import Dict, Iterator

from opentelemetry import trace as otel_trace
from opentelemetry.trace import Span, SpanKind

@dataclass
class BusAttrs:
    topic: str
    schema_fqdn: str
    message_id: str
    partition_key: str
    run_id: str

@contextmanager
def start_span(name: str, kind: SpanKind = SpanKind.INTERNAL, **attrs) -> Iterator[Span]:
    tracer = otel_trace.get_tracer("ampyobs")
    with tracer.start_as_current_span(name, kind=kind) as span:
        for k, v in attrs.items():
            span.set_attribute(k, v)
        yield span  # span will end on exiting the 'with'

@contextmanager
def start_bus_publish(a: BusAttrs) -> Iterator[Span]:
    with start_span(
        "bus.publish",
        SpanKind.PRODUCER,
        topic=a.topic,
        schema_fqdn=a.schema_fqdn,
        message_id=a.message_id,
        partition_key=a.partition_key,
        run_id=a.run_id,
    ) as span:
        yield span

@contextmanager
def start_bus_consume(headers: Dict[str, str], a: BusAttrs) -> Iterator[Span]:
    from .propagation import extract_trace
    remote_ctx = extract_trace(headers)
    tracer = otel_trace.get_tracer("ampyobs")
    with tracer.start_as_current_span("bus.consume", context=remote_ctx, kind=SpanKind.CONSUMER) as span:
        for k, v in dict(topic=a.topic, schema_fqdn=a.schema_fqdn, message_id=a.message_id, partition_key=a.partition_key, run_id=a.run_id).items():
            span.set_attribute(k, v)
        yield span
